if not MergingChests then MergingChests = { } end

require("lookup")

MergingChests.MergableChestId = MergingChests.MergableChestNameToId[settings.startup["mergable-chest-name"].value]

-- globaly used functions
function string.starts(String, Start)
   return string.sub(String, 1, string.len(Start)) == Start
end

function math.round(num)
	if num >= 0 then
		return math.floor(num + 0.5)
	else
		return math.ceil(num - 0.5)
	end
end

function MergingChests.GetChestSize(entity)
	if entity.name == MergingChests.MergableChestId then
		return 1, 1
	elseif string.starts(entity.name, "wide-chest-") then
		return tonumber(string.sub(entity.name, 12)), 1
	elseif string.starts(entity.name, "high-chest-") then
		return 1, tonumber(string.sub(entity.name, 12))
	else
		return 0, 0
	end
end

function MergingChests.MoveInventories(from, to)
	local j = 1
	local l = 1
	local toInventory = to[j].get_inventory(1)
	local bar = 0
	
	for _, entity in ipairs(from) do
		bar = bar + entity.get_inventory(1).getbar()
	end
	
	for i = 1, #from do
		local fromInventory = from[i].get_inventory(1)
		
		for k = 1, #fromInventory do
			if fromInventory[k].valid_for_read then
				toInventory[l].set_stack(fromInventory[k])
				l = l + 1
				
				if l > #toInventory then
					j = j + 1
					
					if j > #to then
						-- should never happen but just to be sure
						goto setbar
					end
					
					toInventory = to[j].get_inventory(1)
					l = 1
				end
			end
		end
	end
	
	::setbar::
	local singleBar = math.round(bar / #to)
	for _, entity in ipairs(to) do
		entity.get_inventory(1).setbar(singleBar)
	end
	to[#to].get_inventory(1).setbar(singleBar + bar - #to * singleBar)
end

function MergingChests.ReconnectCircuits(from, to)
	local fromSet = { }
	for _, entity in ipairs(from) do
		fromSet[entity] = entity
	end
	
	local connections = { }
	local red = false
	local green = false
	for _, entity in ipairs(from) do
		for __, connection in ipairs(entity.circuit_connection_definitions) do
			if not fromSet[connection.target_entity] then
				table.insert(connections, connection)
				
				red = red or connection.wire == defines.wire_type.red
				green = green or connection.wire == defines.wire_type.green
			end
		end
	end
	
	if #connections > 0 then
		--connect all "to" entities together
		for i = 1, #to - 1 do
			if red then
				to[i].connect_neighbour{wire = defines.wire_type.red, target_entity = to[i + 1]}
			end
			
			if green then
				to[i].connect_neighbour{wire = defines.wire_type.green, target_entity = to[i + 1]}
			end
		end
		
		for _, connection in ipairs(connections) do
			local closestEntity = nil
			local min = nil
			
			for __, entity in ipairs(to) do
				local diffX = entity.position.x - connection.target_entity.position.x
				local diffY = entity.position.y - connection.target_entity.position.y
				
				if not min or (diffX * diffX + diffY * diffY < min) then
					min = diffX * diffX + diffY * diffY
					closestEntity = entity
				end
			end
			
			closestEntity.connect_neighbour(connection)
		end
	end
end

-- merging functions
function MergingChests.FindChestsBounds(entities)
	local minX, minY, maxX, maxY = nil, nil, nil, nil
	
	for _, entity in ipairs(entities) do
		local width, height = MergingChests.GetChestSize(entity)
		local offsetX = width / 2
		local offsetY = height / 2
		
		if not minX or (minX > entity.position.x - offsetX) then minX = entity.position.x - offsetX end
		if not minY or (minY > entity.position.y - offsetY) then minY = entity.position.y - offsetY end
		if not maxX or (maxX < entity.position.x + offsetX) then maxX = entity.position.x + offsetX end
		if not maxY or (maxY < entity.position.y + offsetY) then maxY = entity.position.y + offsetY end
	end
	
	return { minX = minX, minY = minY, maxX = maxX - 1, maxY = maxY - 1 }
end

function MergingChests.SortIntoGroups(entities)
	local bounds = MergingChests.FindChestsBounds(entities, FindChestArea)
	
	local mapH = { }
	local mapV = { }
	
	-- init maps
	for x = bounds.minX, bounds.maxX do
		mapV[x] = { }
	end
	for y = bounds.minY, bounds.maxY do
		mapH[y] = { }
	end
	
	-- fill maps
	for _, entity in ipairs(entities) do
		local width, height = MergingChests.GetChestSize(entity)
		width = width - 1
		height = height - 1
		
		if width > height then
			-- horizontal chest
			for i = entity.position.x - width / 2, entity.position.x + width / 2 do
				mapH[math.floor(entity.position.y)][math.floor(i)] = entity
			end
			
		elseif height > width then
			-- vertical chest
			for i = entity.position.y - height / 2, entity.position.y + height / 2 do
				mapV[math.floor(entity.position.x)][math.floor(i)] = entity
			end
			
		elseif width == 0 and height == 0 then
			-- normal chest
			mapH[math.floor(entity.position.y)][math.floor(entity.position.x)] = entity
			mapV[math.floor(entity.position.x)][math.floor(entity.position.y)] = entity
		end
	end
	
	local groups = { }
	local merged = false
	
	repeat
		merged = false
		
		local yH, xH, lengthH = MergingChests.FindLargestChest(mapH, bounds.minY, bounds.maxY, bounds.minX, bounds.maxX)
		local xV, yV, lengthV = MergingChests.FindLargestChest(mapV, bounds.minX, bounds.maxX, bounds.minY, bounds.maxY)
		
		local size = 0
		local newGroup = { }
		
		if lengthH > 0 and lengthH >= lengthV then -- merge horizontal
			
			for x = xH, xH + lengthH do
				if #newGroup == 0 or newGroup[#newGroup] ~= mapH[yH][x] then-- first chest or not last added chest
					local s = math.max(MergingChests.GetChestSize(mapH[yH][x]))
					
					if size + s > settings.startup["max-chest-size"].value then
						break
					end
					
					table.insert(newGroup, mapH[yH][x])
					size = size + s
				end
				
				mapH[yH][x] = nil
				mapV[x][yH] = nil
			end
			
			if size > 1 and #newGroup > 1 then
				table.insert(groups, { horizontal = true, entities = newGroup })
				merged = true
			end
			
		elseif lengthV > 0 then -- merge vertical
			
			for y = yV, yV + lengthV do
				if #newGroup == 0 or newGroup[#newGroup] ~= mapV[xV][y] then-- first chest or not last added chest
					local s = math.max(MergingChests.GetChestSize(mapV[xV][y]))
					
					if size + s > settings.startup["max-chest-size"].value then
						break
					end
					
					table.insert(newGroup, mapV[xV][y])
					size = size + s
				end
				
				mapH[y][xV] = nil
				mapV[xV][y] = nil
			end
			
			if size > 1 and #newGroup > 1 then
				table.insert(groups, { horizontal = false, entities = newGroup })
				merged = true
			end
		end
		
	until not merged
	
	return groups
end

function MergingChests.FindLargestChest(map, firstStart, firstEnd, secondStart, secondEnd)
	local length = 0
	local indexX = -1
	local indexY = -1
	
	local lengths = { }
	
	for x = firstStart, firstEnd do
		for y = secondStart + 1, secondEnd do
			if map[x][y] and map[x][y - 1] then
				-- continue row
				lengths[y] = (lengths[y - 1] or 0) + 1
			else
				-- reset
				lengths[y] = 0
			end
			
			if length < lengths[y] then
				length = lengths[y]
				indexX = x
				indexY = y - lengths[y]
			end
		end
	end
	
	return indexX, indexY, length
end

function MergingChests.CalculateMergePosition(entities)
	local W, H = 0, 0
	local position = { x = 0, y = 0 }
	for __, entity in ipairs(entities) do
		width, height = MergingChests.GetChestSize(entity)
		W, H = W + width, H + height
		position.x = position.x + entity.position.x * width
		position.y = position.y + entity.position.y * height
	end
	
	position.x = position.x / W
	position.y = position.y / H
	
	return position, math.max(W, H)
end

function MergingChests.CreateMergeHorizontaly(entities, player)
	local position, size = MergingChests.CalculateMergePosition(entities)
	
	return player.surface.create_entity{name = "wide-chest-"..size, position = position, force = player.force, player = player }
end

function MergingChests.CreateMergeVerticaly(entities, player)
	local position, size = MergingChests.CalculateMergePosition(entities)
	
	return player.surface.create_entity{name = "high-chest-"..size, position = position, force = player.force, player = player }
end

-- splitting functions
function MergingChests.CreateSplitHorizontaly(entity, player, count)
	local newEntities = { }
	for i = 1, count do
		table.insert(newEntities, player.surface.create_entity{name = MergingChests.MergableChestId, position = {entity.position.x - (count + 1) / 2 + i, entity.position.y}, force = player.force, player = player})
	end
	
	return newEntities
end

function MergingChests.CreateSplitVerticaly(entity, player, count)
	local newEntities = { }
	for i = 1, count do
		table.insert(newEntities, player.surface.create_entity{name = MergingChests.MergableChestId, position = {entity.position.x, entity.position.y - (count + 1) / 2 + i}, force = player.force, player = player})
	end
	
	return newEntities
end

-- merging of chests
function MergingChests.OnPlayerSelectedArea(event)
	if event.item and event.item == "merge-chest-selector" then
		local player = game.players[event.player_index]
		
		-- use event entities and remove everything but mergable chests
		local entities = event.entities
		for i = #entities, 1, -1 do
			if math.max(MergingChests.GetChestSize(entities[i])) == 0 then
				table.remove(entities, i)
			end
		end
		
		if #entities > 1 then
			for _, group in ipairs(MergingChests.SortIntoGroups(entities)) do
				if #group.entities > 1 then
					local newChest
					if group.horizontal then
						newChest = MergingChests.CreateMergeHorizontaly(group.entities, player)
					else
						newChest = MergingChests.CreateMergeVerticaly(group.entities, player)
					end
					
					MergingChests.MoveInventories(group.entities, { newChest })
					MergingChests.ReconnectCircuits(group.entities, { newChest })
					
					for _, entity in ipairs(group.entities) do
						entity.destroy()
					end
				end
			end
		end
	end
end

-- splitting of chests
function MergingChests.OnPlayerAltSelectedArea(event)
	if event.item and event.item == "merge-chest-selector" then
		local player = game.players[event.player_index]
		
		-- use event entities and remove everything but wide-chests and high-chests
		local entities = event.entities
		for i = #entities, 1, -1 do
			if math.max(MergingChests.GetChestSize(entities[i])) < 2 then
				table.remove(entities, i)
			end
		end
		
		for _, entity in ipairs(entities) do
			local width, height = MergingChests.GetChestSize(entity)
			local newEntities
			
			if width > 1 then
				newEntities = MergingChests.CreateSplitHorizontaly(entity, player, width)
			elseif height > 1 then
				newEntities = MergingChests.CreateSplitVerticaly(entity, player, height)
			end
			
			MergingChests.MoveInventories({ entity }, newEntities)
			MergingChests.ReconnectCircuits({ entity }, newEntities)
			
			entity.destroy()
		end
	end
end

script.on_event(defines.events.on_player_selected_area, MergingChests.OnPlayerSelectedArea)
script.on_event(defines.events.on_player_alt_selected_area, MergingChests.OnPlayerAltSelectedArea)