MergingChests.MergableChestInventory = data.raw["container"][MergingChests.MergableChestId].inventory_size

local segmentsData = MergingChests.SpriteSegmentsData[MergingChests.MergableChestId]

function MergingChests.CreateChestEntities(size)
	local inventorySize = MergingChests.MergableChestInventory * size
	
	inventorySize = inventorySize * settings.startup["inventory-size-multiplier"].value
	if inventorySize > settings.startup["inventory-size-limit"].value then
		inventorySize = settings.startup["inventory-size-limit"].value
	end
	
	-- prepare sprite layers
	local spriteH = { }
	for i = (size % 2) / 2, (size - 1) / 2 do
		table.insert(spriteH,
		{
			filename = segmentsData.entity_sprite,
			priority = "high",
			x = segmentsData.wide_segments.middle.x,
			y = segmentsData.wide_segments.middle.y,
			width = segmentsData.wide_segments.middle.width,
			height = segmentsData.wide_segments.middle.height,
			shift =
			{
				i + segmentsData.shift.x / 32,
				segmentsData.shift.y / 32
			},
		})
		table.insert(spriteH,
		{
			filename = segmentsData.entity_sprite,
			priority = "high",
			x = segmentsData.wide_segments.middle.x,
			y = segmentsData.wide_segments.middle.y,
			width = segmentsData.wide_segments.middle.width,
			height = segmentsData.wide_segments.middle.height,
			shift =
			{
				-i + segmentsData.shift.x / 32,
				segmentsData.shift.y / 32
			},
		})
	end
	table.insert(spriteH,
	{
		filename = segmentsData.entity_sprite,
		priority = "high",
		x = segmentsData.wide_segments.left.x,
		y = segmentsData.wide_segments.left.y,
		width = segmentsData.wide_segments.left.width,
		height = segmentsData.wide_segments.left.height,
		shift =
		{
			-size / 2 + (segmentsData.wide_segments.left.width / 2 + segmentsData.shift.x) / 32,
			segmentsData.shift.y / 32
		}
	})
	table.insert(spriteH,
	{
		filename = segmentsData.entity_sprite,
		priority = "high",
		x = segmentsData.wide_segments.right.x,
		y = segmentsData.wide_segments.right.y,
		width = segmentsData.wide_segments.right.width,
		height = segmentsData.wide_segments.right.height,
		shift =
		{
			size / 2 - (segmentsData.wide_segments.right.width / 2 - segmentsData.shift.x) / 32,
			segmentsData.shift.y / 32
		}
	})
	table.insert(spriteH,
	{
		filename = segmentsData.shadow_sprite,
		priority = "high",
		x = segmentsData.wide_segments.shadow.x,
		y = segmentsData.wide_segments.shadow.y,
		width = segmentsData.wide_segments.shadow.width,
		height = segmentsData.wide_segments.shadow.height,
		shift =
		{
			size / 2 + (segmentsData.wide_segments.shadow.width / 2 + segmentsData.shift.x) / 32,
			segmentsData.shift.y / 32
		},
		draw_as_shadow = true,
	})
	local spriteV =
	{
		{
			filename = segmentsData.entity_sprite,
			priority = "high",
			x = segmentsData.high_segments.top.x,
			y = segmentsData.high_segments.top.y,
			width = segmentsData.high_segments.top.width,
			height = segmentsData.high_segments.top.height,
			shift =
			{
				segmentsData.shift.x / 32,
				-size / 2 + (segmentsData.high_segments.top.height / 2 + segmentsData.shift.y) / 32
			}
		},
		{
			filename = segmentsData.shadow_sprite,
			priority = "high",
			x = segmentsData.high_segments.shadow_top.x,
			y = segmentsData.high_segments.shadow_top.y,
			width = segmentsData.high_segments.shadow_top.width,
			height = segmentsData.high_segments.shadow_top.height,
			shift =
			{
				(16 + segmentsData.high_segments.shadow_top.width / 2 + segmentsData.shift.x) / 32,
				-size / 2 + (segmentsData.high_segments.shadow_top.y + segmentsData.high_segments.shadow_top.height / 2 + segmentsData.shift.y) / 32
			},
			draw_as_shadow = true,
		},
		{
			filename = segmentsData.entity_sprite,
			priority = "high",
			x = segmentsData.high_segments.bottom.x,
			y = segmentsData.high_segments.bottom.y,
			width = segmentsData.high_segments.bottom.width,
			height = segmentsData.high_segments.bottom.height,
			shift =
			{
				segmentsData.shift.x / 32,
				size / 2 - (segmentsData.high_segments.bottom.height / 2 + segmentsData.shift.y) / 32
			}
		},
		{
			filename = segmentsData.shadow_sprite,
			priority = "high",
			x = segmentsData.high_segments.shadow_bottom.x,
			y = segmentsData.high_segments.shadow_bottom.y,
			width = segmentsData.high_segments.shadow_bottom.width,
			height = segmentsData.high_segments.shadow_bottom.height,
			shift =
			{
				(16 + segmentsData.high_segments.shadow_bottom.width / 2 + segmentsData.shift.x) / 32,
				size / 2 - (segmentsData.high_segments.shadow_bottom.height / 2 + segmentsData.shift.y) / 32
			},
			draw_as_shadow = true,
		}
	}
	for i = 0, size - 2 do
		table.insert(spriteV,
		{
			filename = segmentsData.entity_sprite,
			priority = "high",
			x = segmentsData.high_segments.middle.x,
			y = segmentsData.high_segments.middle.y,
			width = segmentsData.high_segments.middle.width,
			height = segmentsData.high_segments.middle.height,
			shift =
			{
				segmentsData.shift.x / 32,
				i - size / 2 + (segmentsData.high_segments.middle.y + segmentsData.high_segments.middle.height / 2 + segmentsData.shift.y) / 32
			}
		})
		table.insert(spriteV,
		{
			filename = segmentsData.shadow_sprite,
			priority = "high",
			x = segmentsData.high_segments.shadow_middle.x,
			y = segmentsData.high_segments.shadow_middle.y,
			width = segmentsData.high_segments.shadow_middle.width,
			height = segmentsData.high_segments.shadow_middle.height,
			shift =
			{
				(16 + segmentsData.high_segments.shadow_middle.width / 2 + segmentsData.shift.x) / 32,
				i - size / 2 + (segmentsData.high_segments.shadow_middle.y + segmentsData.high_segments.shadow_middle.height / 2 + segmentsData.shift.y) / 32
			},
			draw_as_shadow = true,
		})
	end
	
	-- prepare circuit connectors
	local connectorH = circuit_connector_definitions["chest"]
	local connectorV = circuit_connector_definitions.create
	(
		universal_connector_template,
		{
			{
				variation = 24,
				main_offset = util.by_pixel(-16, 0),
				shadow_offset = util.by_pixel(7.5, 7.5),
				show_shadow = false,
			},
		}
	)
	
	return
	{
		{
			type = "container",
			name = "wide-chest-"..size,
			localised_name = { "chest-name.wide-chest", size },
			icon = "__base__/graphics/icons/"..MergingChests.MergableChestId..".png",
			icon_size = 32,
			flags = { "placeable-neutral", "player-creation" },
			minable = { mining_time = 2, result = MergingChests.MergableChestId, count = size },
			placeable_by = { item = MergingChests.MergableChestId, count = size },
			max_health = 350,
			corpse = "medium-remnants",
			dying_explosion = "medium-explosion",
			open_sound = { filename = "__base__/sound/machine-open.ogg", volume = 0.85 },
			close_sound = { filename = "__base__/sound/machine-close.ogg", volume = 0.75 },
			vehicle_impact_sound = { filename = "__base__/sound/car-metal-impact.ogg", volume = 0.65 },
			collision_box = { { -size / 2 + 0.15, -0.35 }, { size / 2 - 0.15, 0.35 } },
			selection_box = { { -size / 2, -0.5 }, { size / 2, 0.5 } },
			fast_replaceable_group = "container",
			inventory_size = inventorySize,
			picture =
			{
				layers = spriteH
			},
			circuit_wire_connection_point = connectorH.points,
			circuit_connector_sprites = connectorH.sprites,
			circuit_wire_max_distance = default_circuit_wire_max_distance,
		},
		{
			type = "container",
			name = "high-chest-"..size,
			localised_name = { "chest-name.high-chest", size },
			icon = "__base__/graphics/icons/"..MergingChests.MergableChestId..".png",
			icon_size = 32,
			flags = { "placeable-neutral", "player-creation" },
			minable = { mining_time = 2, result = MergingChests.MergableChestId, count = size },
			placeable_by = { item = MergingChests.MergableChestId , count = size },
			max_health = 350,
			corpse = "medium-remnants",
			dying_explosion = "medium-explosion",
			open_sound = { filename = "__base__/sound/machine-open.ogg", volume = 0.85 },
			close_sound = { filename = "__base__/sound/machine-close.ogg", volume = 0.75 },
			vehicle_impact_sound = { filename = "__base__/sound/car-metal-impact.ogg", volume = 0.65 },
			collision_box = { { -0.35, -size / 2 + 0.15 }, { 0.35, size / 2 - 0.15 } },
			selection_box = { { -0.5, -size / 2 }, { 0.5, size / 2 } },
			fast_replaceable_group = "container",
			inventory_size = inventorySize,
			picture =
			{
				layers = spriteV
			},
			circuit_wire_connection_point = connectorV.points,
			circuit_connector_sprites = connectorV.sprites,
			circuit_wire_max_distance = default_circuit_wire_max_distance,
		}
	}
end

for i = 2, settings.startup["max-chest-size"].value do
	data:extend(MergingChests.CreateChestEntities(i))
end